# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import csv
import tempfile
from pathlib import Path

import pytest

from .manifest import list_videos_in_manifest, VideoMetadata
from .universities import BUCKET_TO_UNIV


def test_constructor():

    row = {
        "video_uid": "12345",
        "canonical_s3_location": "s3://ego4d-georgiatech/object/key/123",
    }

    metadata = VideoMetadata(row=row)

    # Known university
    assert metadata.uid == "12345"
    assert metadata.s3_path == "s3://ego4d-georgiatech/object/key/123"
    assert metadata.s3_bucket == "ego4d-georgiatech"
    assert metadata.s3_object_key == "object/key/123"
    assert metadata.university == BUCKET_TO_UNIV["ego4d-georgiatech"]

    # Unknown university
    row["canonical_s3_location"] = "s3://unrecognized-university/object/key/123"
    # Metadata should copy the row
    assert (
        metadata.raw_data["canonical_s3_location"]
        == "s3://ego4d-georgiatech/object/key/123"
    )

    unknown_univ = VideoMetadata(row=row)
    assert unknown_univ.s3_path == "s3://unrecognized-university/object/key/123"
    assert unknown_univ.s3_bucket == "unrecognized-university"
    assert unknown_univ.s3_object_key == "object/key/123"
    assert not unknown_univ.university


def test_csv_reader():

    with tempfile.NamedTemporaryFile(mode="w") as f:

        writer = csv.DictWriter(f, fieldnames=["video_uid", "canonical_s3_location"])
        writer.writeheader()
        writer.writerow(
            {
                "video_uid": "12345",
                "canonical_s3_location": "s3://ego4d-georgiatech/object/key/123",
            }
        )
        writer.writerow(
            {
                "video_uid": "7890",
                "canonical_s3_location": "s3://unrecognized-university/other/key/456",
            }
        )
        f.flush()

        full_generator = list_videos_in_manifest(
            manifest_path=Path(f.name), for_universities=set(), benchmarks=set()
        )
        georgiatech_generator = list_videos_in_manifest(
            manifest_path=Path(f.name),
            for_universities={BUCKET_TO_UNIV["ego4d-georgiatech"]},
            benchmarks=set(),
        )

        # Full generator should return both videos
        metadata = next(full_generator)
        assert metadata.uid == "12345"
        assert metadata.s3_path == "s3://ego4d-georgiatech/object/key/123"
        assert metadata.s3_bucket == "ego4d-georgiatech"
        assert metadata.s3_object_key == "object/key/123"
        assert metadata.university == BUCKET_TO_UNIV["ego4d-georgiatech"]

        metadata = next(full_generator)
        assert metadata.uid == "7890"
        assert metadata.s3_path == "s3://unrecognized-university/other/key/456"
        assert metadata.s3_bucket == "unrecognized-university"
        assert metadata.s3_object_key == "other/key/456"
        assert not metadata.university

        with pytest.raises(StopIteration):
            next(full_generator)

        # Georgia Tech should only return one video
        metadata = next(georgiatech_generator)
        assert metadata.uid == "12345"
        assert metadata.s3_path == "s3://ego4d-georgiatech/object/key/123"
        assert metadata.s3_bucket == "ego4d-georgiatech"
        assert metadata.s3_object_key == "object/key/123"
        assert metadata.university == BUCKET_TO_UNIV["ego4d-georgiatech"]

        with pytest.raises(StopIteration):
            next(georgiatech_generator)


def test_csv_benchmarks():
    with tempfile.NamedTemporaryFile(mode="w") as f:
        writer = csv.DictWriter(
            f, fieldnames=["video_uid", "canonical_s3_location", "benchmarks"]
        )
        writer.writeheader()
        writer.writerow(
            {
                "video_uid": "12345",
                "canonical_s3_location": "s3://ego4d-georgiatech/object/key/123",
                "benchmarks": "[FHO |EM]",
            }
        )
        writer.writerow(
            {
                "video_uid": "7890",
                "canonical_s3_location": "s3://unrecognized-university/other/key/456",
                "benchmarks": "",
            }
        )
        writer.writerow(
            {
                "video_uid": "6789",
                "canonical_s3_location": "s3://unrecognized-university/other/key/456",
                "benchmarks": "[]",
            }
        )
        writer.writerow(
            {
                "video_uid": "23456",
                "canonical_s3_location": "s3://unrecognized-university/other/key/456",
                "benchmarks": "[ EM]",
            }
        )
        f.flush()

        full_generator = list_videos_in_manifest(
            manifest_path=Path(f.name), for_universities=set(), benchmarks=set()
        )
        em_generator = list_videos_in_manifest(
            manifest_path=Path(f.name),
            for_universities=set(),
            benchmarks=set(["em"]),
        )

        # Full generator should return both videos
        metadata = next(full_generator)
        assert metadata.uid == "12345"
        assert metadata.s3_path == "s3://ego4d-georgiatech/object/key/123"
        assert metadata.s3_bucket == "ego4d-georgiatech"
        assert metadata.s3_object_key == "object/key/123"
        assert metadata.university == BUCKET_TO_UNIV["ego4d-georgiatech"]

        metadata = next(full_generator)
        assert metadata.uid == "7890"
        assert metadata.s3_path == "s3://unrecognized-university/other/key/456"
        assert metadata.s3_bucket == "unrecognized-university"
        assert metadata.s3_object_key == "other/key/456"
        assert not metadata.university

        metadata = next(full_generator)
        assert metadata.uid == "6789"

        metadata = next(full_generator)
        assert metadata.uid == "23456"

        with pytest.raises(StopIteration):
            next(full_generator)

        # EM should only return two videos
        metadata = next(em_generator)
        assert metadata.uid == "12345"

        metadata = next(em_generator)
        assert metadata.uid == "23456"

        with pytest.raises(StopIteration):
            next(em_generator)



import math
from typing import Optional

import av

import torch
from pytorchvideo.transforms import Normalize, ShortSideScale
from torchvision.transforms._transforms_video import CenterCropVideo

try:
    from torchaudio.io import StreamReader
except ImportError:
    print("WARN: torchaudio not installed", flush=True)


def get_video_meta(path):
    with av.open(path) as cont:
        n_frames = cont.streams[0].frames
        codec = cont.streams[0].codec.name
        codec_long_name = cont.streams[0].codec.long_name
        tb = cont.streams[0].time_base

        all_pts = []
        for x in cont.demux(video=0):
            if x.pts is None:
                continue
            all_pts.append(x.pts)

        assert len(all_pts) == n_frames
        return {
            "all_pts": sorted(all_pts),
            "codec": codec,
            "codec_long_name": codec_long_name,
            "tb": tb,
            "width": cont.streams.video[0].width,
            "height": cont.streams.video[0].height,
        }


def _yuv_to_rgb(img: torch.Tensor) -> torch.Tensor:
    img = img.to(torch.float)
    y = img[..., 0, :, :]
    u = img[..., 1, :, :]
    v = img[..., 2, :, :]

    y /= 255
    u = u / 255 - 0.5
    v = v / 255 - 0.5

    r = y + 1.14 * v
    g = y + -0.396 * u - 0.581 * v
    b = y + 2.029 * u

    rgb = torch.stack([r, g, b], -1)
    return rgb.permute(3, 0, 1, 2)


def _derive_cthw_axis_order(axis_order):
    mapping = {"c": 0, "t": 1, "h": 2, "w": 3}
    assert set(mapping.keys()) == set(
        axis_order
    ), "please provide 't', 'c', 'h', 'w' in any order"
    assert len(axis_order) == 4, "please provide 't', 'c', 'h', 'w' in any order"
    result = []
    for ch in axis_order:
        result.append(mapping[ch])
    return tuple(result)


class StridedReader:
    def __init__(self, path, stride, frame_window_size, axis_order: str = "cthw"):
        self.path = path
        self.meta = get_video_meta(path)
        self.all_pts = self.meta["all_pts"]
        self.stride = stride
        self.frame_window_size = frame_window_size
        self.axis_order = axis_order
        self.cthw_to_axis_order = _derive_cthw_axis_order(axis_order)
        if self.stride == 0:
            self.stride = self.frame_window_size

    def __getitem__(self, _: int) -> torch.Tensor:
        raise AssertionError("Not implemented")

    def __len__(self):
        return int(
            math.ceil((len(self.all_pts) - self.frame_window_size) / self.stride)
        )


class TorchAudioStreamReader(StridedReader):
    def __init__(
        self,
        path: str,
        resize: Optional[int],
        crop: Optional[int],
        mean: Optional[torch.Tensor],
        std: Optional[torch.Tensor],
        frame_window_size: int,
        stride: int,
        gpu_idx: int,
        axis_order: str = "cthw",
        uint8_scale: bool = False,
        resize_with_hardware: Optional[int] = None,
    ):
        """
        NOTE:
        - resize_on_hardware uses CUVID's hardware resize operation. I don't
          know what interpolation algorithm it uses (can't find documentation).
          Use with caution.
        """
        super().__init__(path, stride, frame_window_size, axis_order)

        self.mean = mean
        self.crop = crop
        self.std = std
        self.resize = resize
        self.resize_transform = (
            ShortSideScale(self.resize) if self.resize is not None else None
        )
        self.norm_transform = (
            Normalize(mean=self.mean, std=self.std) if self.mean is not None else None
        )
        self.crop_transform = (
            CenterCropVideo(self.crop) if self.crop is not None else None
        )
        self.resize_on_hardware = resize_with_hardware
        self.uint8_scale = uint8_scale
        self.create_underlying_cont(gpu_idx)

    def create_underlying_cont(self, gpu_id):
        self.gpu_id = gpu_id

        decoder_basename = self.meta["codec"]
        if self.resize_on_hardware:
            assert self.gpu_id >= 0
            w, h = self.meta["width"], self.meta["height"]
            roh = self.resize_on_hardware
            rw, rh = roh, roh
            if w < h:
                rh = int(math.floor((float(h) / w) * roh))
            elif w > h:
                rw = int(math.floor((float(w) / h) * roh))

            decoder_opt = (
                {"resize": f"{rw}x{rh}", "gpu": f"{gpu_id}"}
                if self.resize is not None
                else {"gpu": f"{gpu_id}"}
            )
            self.conf = {
                "decoder": f"{decoder_basename}_cuvid",
                "hw_accel": f"cuda:{gpu_id}",
                "decoder_option": decoder_opt,
                "stream_index": 0,
            }
        else:
            self.conf = {
                "decoder": decoder_basename,
                "stream_index": 0,
            }

        self.resize_transform = (
            ShortSideScale(self.resize) if self.resize is not None else None
        )

        self.cont = StreamReader(self.path)
        self.cont.add_video_stream(self.frame_window_size, **self.conf)
        self.cont.fill_buffer()

    def __getitem__(self, idx: int) -> dict:
        frame_i = self.stride * idx
        frame_j = frame_i + self.frame_window_size - 1
        assert frame_i >= 0 and frame_j < len(self.all_pts)

        frame_i_pts = self.all_pts[frame_i]
        self.cont.seek(float(frame_i_pts * self.meta["tb"]))
        fs = None
        for fs in self.cont.stream():
            break

        assert fs is not None
        assert len(fs) == 1
        ret = _yuv_to_rgb(fs[0])  # TODO: optimize me
        assert ret.shape[1] == self.frame_window_size
        if self.resize_transform is not None:
            ret = self.resize_transform(ret)
        if self.crop_transform is not None:
            ret = self.crop_transform(ret)
        if self.norm_transform is not None:
            ret = self.norm_transform(ret)
        ret = ret.permute(self.cthw_to_axis_order)
        if self.uint8_scale:
            ret *= 255
            ret = ret.to(dtype=torch.uint8)
        return {
            "video": ret,
            "frame_start_idx": frame_i,
            "frame_end_idx": frame_j,
        }


class PyAvReader(StridedReader):
    def __init__(
        self,
        path: str,
        resize: Optional[int],
        crop: Optional[int],
        mean: Optional[torch.Tensor],
        std: Optional[torch.Tensor],
        frame_window_size: int,
        stride: int,
        gpu_idx: int,
        axis_order: str = "cthw",
        uint8_scale: bool = False,
    ):
        super().__init__(path, stride, frame_window_size, axis_order)

        if gpu_idx >= 0:
            print("WARN: GPU decoding not supported for PyAV, using CPU")

        self.mean = mean
        self.crop = crop
        self.std = std
        self.resize = resize
        self.resize_transform = (
            ShortSideScale(self.resize) if self.resize is not None else None
        )
        self.norm_transform = (
            Normalize(mean=self.mean, std=self.std) if self.mean is not None else None
        )
        self.crop_transform = (
            CenterCropVideo(self.crop) if self.crop is not None else None
        )
        self.uint8_scale = uint8_scale
        self.create_underlying_cont(gpu_idx)

    def create_underlying_cont(self, _):
        self.cont = av.open(self.path)

    def __getitem__(self, idx: int) -> dict:
        frame_i = self.stride * idx
        frame_j = frame_i + self.frame_window_size
        assert frame_i >= 0 and frame_j < len(self.all_pts)

        frame_i_pts = self.all_pts[frame_i]
        frame_j_pts = self.all_pts[frame_j]
        self.cont.seek(frame_i_pts, stream=self.cont.streams.video[0])
        fs = []
        for f in self.cont.decode(video=0):
            if f.pts < frame_i_pts:
                continue
            if f.pts >= frame_j_pts:
                break
            fs.append(
                (
                    f.pts,
                    torch.tensor(f.to_ndarray(format="rgb24"), dtype=torch.float32)
                    / 255,
                )
            )
        fs = sorted(fs, key=lambda x: x[0])
        # NOTE: channel first for the transforms
        ret = torch.stack([x[1] for x in fs]).permute(3, 0, 1, 2)
        if self.resize_transform is not None:
            ret = self.resize_transform(ret)
        if self.crop_transform is not None:
            ret = self.crop_transform(ret)
        if self.norm_transform is not None:
            ret = self.norm_transform(ret)
        ret = ret.permute(self.cthw_to_axis_order)
        if self.uint8_scale:
            ret *= 255
            ret = ret.to(dtype=torch.uint8)
        return {
            "video": ret,
            "frame_start_idx": frame_i,
            "frame_end_idx": frame_j - 1,
        }



import functools
import os

import h5py
import pandas as pd
import torch
from ego4d.features.config import load_model
from ego4d.research.clep.config import EgoCharadePreprocessConfig, TrainConfig
from ego4d.research.clep.preprocess.common import (
    get_language_model,
    run_feature_extraction,
)

from ego4d.research.common import batch_it, create_executor
from omegaconf import OmegaConf
from tqdm.auto import tqdm


def preprocess_ego_charade(
    config: TrainConfig, char_config: EgoCharadePreprocessConfig
):
    out_dir = config.pre_config.root_dir
    os.makedirs(out_dir, exist_ok=True)

    df = pd.read_csv(char_config.set_path)

    root_path = char_config.video_root_path
    feature_extract_config = OmegaConf.load(
        config.input_config.feature_extract_config_path
    )

    class_desc_path = char_config.class_desc_path
    class_name_df = pd.read_csv(class_desc_path, header=None)
    class_names = [" ".join(x[1].split(" ")[1:]) for x in class_name_df.itertuples()]

    def get_label_name(x):
        x.replace("Someone", "")
        x.replace("is", "")
        return x.lower()

    sentences_ego = [
        f"Camera wearer is {get_label_name(clazz)}" for clazz in class_names
    ]

    sentences_non_ego = [
        f"The person in this video is {get_label_name(clazz)}" for clazz in class_names
    ]
    model = get_language_model(config)
    # pyre-ignore
    label_name_fv = model.encode(
        class_names,
        device="cuda",
        show_progress_bar=True,
    )
    # pyre-ignore
    sent_ego_fv = model.encode(
        sentences_ego,
        device="cuda",
        show_progress_bar=True,
    )
    # pyre-ignore
    sent_non_ego = model.encode(
        sentences_non_ego,
        device="cuda",
        show_progress_bar=True,
    )
    torch.save(
        {
            "labels": label_name_fv,
            "sent_ego_fv": sent_ego_fv,
            "sent_non_ego_fv": sent_non_ego,
        },
        os.path.join(out_dir, char_config.out_label_path),
    )
    video_path_ids = [
        (os.path.join(root_path, f"{row.id}.mp4"), row.id) for row in df.itertuples()
    ]
    video_path_ids = [vp for vp in video_path_ids if os.path.exists(vp[0])]

    batches = batch_it(video_path_ids, char_config.num_vids_per_machine)
    executor = create_executor(config.pre_config.slurm_config, len(batches))
    map_fn = functools.partial(
        _preprocess_ego_charade,
        feature_extract_config=feature_extract_config,
    )

    jobs = executor.map_array(map_fn, batches)

    out_path = os.path.join(out_dir, char_config.out_path)
    with h5py.File(out_path, "w") as out_f:
        for j in tqdm(jobs):
            feat = j.result()
            for uid, ret in feat.items():
                out_f.create_dataset(uid, data=ret["features"].numpy())


def _preprocess_ego_charade(video_path_ids, feature_extract_config):
    model = load_model(feature_extract_config, patch_final_layer=True)

    ret = {}
    for path, uid in tqdm(video_path_ids):
        predictions = run_feature_extraction(path, model, feature_extract_config)
        assert predictions is not None
        ret[uid] = {
            "features": predictions.result[path],
        }
    return ret



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import gc
import logging
import math
import os
import time
from collections import defaultdict
from dataclasses import dataclass
from fractions import Fraction
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple, Union

import av
import hydra
import torch
import torchvision
from ego4d.features.config import (
    FeatureExtractConfig,
    get_videos,
    InferenceConfig,
    load_model,
    Video,
)
from ego4d.features.dataset import create_data_loader_or_dset
from omegaconf import OmegaConf
from torch.nn import Module
from tqdm.auto import tqdm


@dataclass
class TimeStats:
    to_load: List[float]
    transfer_device: List[float]
    forward_pass: List[float]

    overall: Optional[float] = None
    to_save: Optional[float] = None


@dataclass
class FeatureExtractionResult:
    result: Dict[str, torch.Tensor]
    time_stats: TimeStats


@dataclass
class ExtractedFeature:
    video_uid: str
    clip_index: int
    start_time_sec: float
    end_time_sec: float
    feature: Union[List[str], torch.Tensor]
    time_to_load: List[float]
    time_transfer_device: List[float]
    time_forward_pass: List[float]


# TODO: use this method in pytorchvideo test_uniform_clip_sampler
# NOTE this is copied from there - hence has been tested
def _num_fvs(
    num_frames: int,
    fps: float,
    stride_frames: int,
    window_size_frames: int,
    backpad_last: bool = True,
) -> int:
    N = num_frames - window_size_frames
    if N < 0:
        return 1

    result = N // stride_frames + 1

    # handle padded frame
    if backpad_last and N % stride_frames != 0:
        result += 1
    return result


def num_fvs(vid: Video, config: InferenceConfig) -> int:
    return _num_fvs(vid.frame_count, config.fps, config.stride, config.frame_window)


def _extract_features(
    model: Module,
    videos: List[Video],
    config: FeatureExtractConfig,
    max_examples: int = -1,
    silent: bool = False,
) -> Iterator[ExtractedFeature]:
    device = config.inference_config.device

    data = create_data_loader_or_dset(videos, config)

    if not silent:
        print(f"extracting from {videos}")

    t1 = time.time()
    for i, x in enumerate(data):
        t2 = time.time()
        load_time = t2 - t1

        t1 = time.time()

        for k in ["video", "audio"]:
            if k not in x or x[k] is None:
                continue

            v = x[k]
            if isinstance(v, list):
                x[k] = [i.to(device) for i in v]
                batch_size = v[0].shape[0]
            else:
                x[k] = v.to(device)
                batch_size = v.shape[0]

        t2 = time.time()
        transfer_time = t2 - t1

        with torch.no_grad():
            t1 = time.time()
            if not config.io.debug_mode:
                fv = model(x)
                if isinstance(fv, torch.Tensor):
                    fv = fv.detach().cpu()
            else:
                vid_inp = x["video"]
                fv = None
                for batch_idx, (video_name, clip_index) in enumerate(
                    zip(x["video_name"], x["clip_index"])
                ):
                    to_dir = f"{config.io.debug_path}/{video_name}"
                    to_path = f"{to_dir}/{clip_index}.jpg"
                    os.makedirs(to_dir, exist_ok=True)
                    grid = torchvision.utils.make_grid(
                        vid_inp[batch_idx].permute(1, 0, 2, 3),
                        nrows=vid_inp.shape[2] // config.inference_config.stride,
                    )
                    torchvision.utils.save_image(grid / 255.0, fp=to_path)  # noqa
            t2 = time.time()

            forward_pass_time = t2 - t1

            yield ExtractedFeature(
                video_uid=x["video_name"],
                clip_index=x["clip_index"],
                start_time_sec=x["clip_start_sec"],
                end_time_sec=x["clip_end_sec"],
                feature=fv,
                time_to_load=load_time,
                time_transfer_device=transfer_time,
                time_forward_pass=forward_pass_time,
            )

        if i % 50 == 0:
            gc.collect()

        assert batch_size is not None
        if max_examples > 0 and (i + 1) * batch_size >= max_examples:
            if not silent:
                print("Breaking...")
            break
        t1 = time.time()


def extract_features(
    videos: List[Video],
    config: FeatureExtractConfig,
    model: Optional[Module] = None,
    log_info: bool = True,
    max_examples: int = -1,
    silent: bool = False,
    assert_feature_size: bool = True,
) -> FeatureExtractionResult:
    if model is None:
        model = load_model(config)

    if log_info and not silent:
        # print(f"config={config}")
        print(f"Number of videos = {len(videos)}")

    fvs = defaultdict(list)
    time_to_load = []
    time_transfer_device = []
    time_forward_pass = []

    uid_to_video_clips = {}
    for v in videos:
        assert v.uid not in uid_to_video_clips
        uid_to_video_clips[v.uid] = v

    # TODO: check this
    # NOTE: not quite right if one is < 0
    batch_size = config.inference_config.batch_size
    total_num_clips = sum(num_fvs(v, config.inference_config) for v in videos)
    total_num_clips /= max(batch_size, 1)
    total_num_clips = math.ceil(total_num_clips)

    if not silent:
        print(
            f"extracting features - there are {total_num_clips} for {len(videos)} videos",
            flush=True,
        )

    for ef in tqdm(
        _extract_features(
            model, videos, config, max_examples=max_examples, silent=silent
        ),
        total=total_num_clips,
    ):
        if config.io.debug_mode:
            continue

        if batch_size == 0:
            key = ef.video_uid
            if isinstance(ef.feature, torch.Tensor):
                ef.feature = ef.feature.cpu().squeeze()
            fvs[key].append(ef)
        else:
            if ef.feature is not None and isinstance(ef.feature, torch.Tensor):
                assert len(ef.video_uid) == len(ef.feature)
                assert len(ef.video_uid) == len(ef.clip_index)

            for i in range(len(ef.video_uid)):
                key = ef.video_uid[i]
                if isinstance(ef.feature, torch.Tensor):
                    ef.feature[i] = ef.feature[i].cpu().squeeze()
                fvs[key].append(ef)

        time_to_load.append(ef.time_to_load)
        time_transfer_device.append(ef.time_transfer_device)
        time_forward_pass.append(ef.time_forward_pass)

    result = {}
    total_num = 0
    for k, efs in fvs.items():
        efs.sort(key=lambda x: x.start_time_sec)

        fv_amount = len(efs)
        if isinstance(efs[0].feature, torch.Tensor):
            if all([e.feature.shape == efs[0].feature.shape for e in efs]):
                result[k] = torch.stack([x.feature for x in efs]).cpu().detach()
            else:
                result[k] = (
                    torch.concat([x.feature for x in efs], dim=1)
                    .cpu()
                    .detach()
                    .squeeze()
                )
        else:
            result[k] = [
                {
                    "start_time_sec": x.start_time_sec.item(),
                    "end_time_sec": x.end_time_sec.item(),
                    "feature": x.feature,
                }
                for x in efs
                if x.feature is not None
            ]

        clip = uid_to_video_clips[k]
        expected_fvs = num_fvs(clip, config.inference_config)
        if config.check_fv_count and expected_fvs != fv_amount:
            if assert_feature_size:
                raise AssertionError(
                    f"{k} should have {expected_fvs} fvs, but has {fv_amount}"
                )

        total_num += fv_amount

    if max_examples > 0:
        assert total_num == max_examples

    return FeatureExtractionResult(
        result=result,
        time_stats=TimeStats(
            to_load=time_to_load,
            transfer_device=time_transfer_device,
            forward_pass=time_forward_pass,
        ),
    )


def perform_feature_extraction(
    videos: List[Video], config: FeatureExtractConfig
) -> TimeStats:
    os.makedirs(config.io.out_path, exist_ok=True)

    # TODO batch this
    time_stats = TimeStats(
        to_load=[],
        transfer_device=[],
        forward_pass=[],
        overall=0,
        to_save=0,
    )

    # sort by smallest video first
    videos.sort(key=lambda x: x.frame_count)

    is_audio_model = config.inference_config.include_audio
    o1 = time.time()
    for vid in tqdm(videos, desc="videos"):
        gc.collect()

        print(f"Extract: {vid}")

        try:
            # TODO: convert this to be a generator
            feature_extract_result = extract_features(
                [vid],
                config,
                assert_feature_size=not is_audio_model,
            )
            result = feature_extract_result.result

            t1 = time.time()
            for k, v in result.items():
                torch.save(v, f"{config.io.out_path}/{k}.pt")
            t2 = time.time()

            time_stats.to_save += t2 - t1
            time_stats.to_load.extend(feature_extract_result.time_stats.to_load)
            time_stats.transfer_device.extend(
                feature_extract_result.time_stats.transfer_device
            )
            time_stats.forward_pass.extend(
                feature_extract_result.time_stats.forward_pass
            )
        except av.error.EOFError as ex:
            logging.exception(f"Failed to extract features for {vid}: {ex}")

    o2 = time.time()

    time_stats.overall = o2 - o1
    return time_stats


@hydra.main(config_path="configs", config_name=None)
def run_extraction(config: FeatureExtractConfig):
    assert (
        config.schedule_config.run_locally
    ), "Local only permitted - use slurm otherwise!"

    print("###################### Feature Extraction Config ####################")
    print(OmegaConf.to_yaml(config))
    print("############################################################")

    # Get uids + {uids -> duration}
    videos, all_videos = get_videos(config)
    print(f"run_extraction: {len(videos)} videos")

    os.makedirs(config.io.out_path, exist_ok=True)
    with open(f"{config.io.out_path}/config.yaml", "w") as out_f:
        out_f.write(OmegaConf.to_yaml(config))

    # print_stats_for_videos(config, all_videos=all_videos, videos=videos)

    result = perform_feature_extraction(videos, config)

    print(f"Extraction Complete: {result}")


if __name__ == "__main__":
    run_extraction()  # pyre-ignore



import bisect
import math
import os
from typing import Any, Callable, Dict, List, Optional, Tuple

import h5py
import torch
from ego4d.research.readers import PyAvReader, StridedReader, TorchAudioStreamReader

from tqdm import tqdm


class LabelledFeatureDset(torch.utils.data.Dataset):
    """
    A simple utility class to load features associated with labels. The input this
    method requires is as follows:
        1. `feature_hdf5_path`: the features transposed to a HDF5 file.
            See `save_features_to_hdf5`
        2. `uid_label_pairs` a list of (uid, label). `label` can be anything
            `uid` is a unique id associated to the `feature_hdf5_path` file.
        3. `aggr_function` a function to aggregate based off given label
    """

    def __init__(
        self,
        feature_hdf5_path: str,
        uid_label_pairs: List[Tuple[str, Any]],
        aggr_function: Optional[Callable[[torch.Tensor, Any], torch.Tensor]] = None,
    ):
        self.uid_label_pairs = uid_label_pairs
        self.features = h5py.File(feature_hdf5_path)
        self.aggr_function = (
            aggr_function
            if aggr_function is not None
            else lambda x, _: torch.tensor(x[0:]).squeeze()
        )
        self.uid_label_pairs = uid_label_pairs
        f_keys = set(self.features.keys())
        l_keys = {uid for uid, _ in self.uid_label_pairs}
        if len(l_keys - f_keys) > 0:
            print(
                f"WARN: missing {len(l_keys - f_keys)} keys in feature hdf5 path: {feature_hdf5_path}"
            )
            self.uid_label_pairs = [
                (uid, label) for uid, label in self.uid_label_pairs if uid in f_keys
            ]

    def __len__(self):
        return len(self.uid_label_pairs)

    def __getitem__(self, idx: int):
        uid, label = self.uid_label_pairs[idx]
        feat = self.aggr_function(self.features[uid], label)
        return feat, label


def save_features_to_hdf5(uids: List[str], feature_dir: str, out_path: str):
    """
    Use this function to preprocess Ego4D features into a HDF5 file with h5py
    """
    with h5py.File(out_path, "w") as out_f:
        for uid in tqdm(uids, leave=True):
            feature_path = os.path.join(feature_dir, f"{uid}.pt")
            fv = torch.load(feature_path)
            out_f.create_dataset(uid, data=fv.numpy())


# Related:
# https://pytorch.org/docs/stable/_modules/torch/utils/data/dataset.html#ConcatDataset
class VideoDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        paths: List[str],
        video_class: type,
        video_class_kwargs: Dict[str, Any],
        max_num_frames_per_video: Optional[int] = None,
        paths_to_n_frames: Optional[Dict[str, int]] = None,
        with_pbar: bool = False,
        transform_fn=None,
        labels_fn=None,
    ):
        paths = sorted(paths)
        self.video_class = video_class
        self.video_class_kwargs = video_class_kwargs
        self.paths = paths
        self.labels_fn = labels_fn
        self.transform_fn = transform_fn
        if paths_to_n_frames is None:
            print("Creating containers")
            path_iter = paths
            if with_pbar:
                path_iter = tqdm(paths)
            self.conts = {
                idx: (p, video_class(p, **video_class_kwargs))
                for idx, p in enumerate(path_iter)
            }
            print("Created containers")

            cont_iter = list(self.conts.values())
            if with_pbar:
                cont_iter = tqdm(cont_iter)

            self.fs_cumsum = [
                (
                    min(len(ct), max_num_frames_per_video)
                    if max_num_frames_per_video
                    else len(ct)
                )
                for _, ct in cont_iter
            ]
            self.fs_cumsum = [0] + torch.cumsum(
                torch.tensor(self.fs_cumsum), dim=0
            ).tolist()
        else:
            self.fs_cumsum = [
                int(
                    math.ceil(
                        (
                            paths_to_n_frames[path]
                            - video_class_kwargs["frame_window_size"]
                        )
                        / video_class_kwargs.get(
                            "stride", video_class_kwargs["frame_window_size"]
                        )
                    )
                )
                for path in paths
            ]
            self.fs_cumsum = [0] + torch.cumsum(
                torch.tensor(self.fs_cumsum), dim=0
            ).tolist()
            self.conts = {idx: (p, None) for idx, p in enumerate(paths)}

    def create_underlying_cont(self, gpu_id):
        # TODO clear_cuda_context_cache()
        self.gpu_id = gpu_id
        for _, c in self.conts.values():
            if c is not None:
                c.create_underlying_cont(self.gpu_id)

    def __getitem__(self, i):
        cont_idx = bisect.bisect_left(self.fs_cumsum, i + 1) - 1
        p, c = self.conts[cont_idx]
        if c is None:
            self.conts[cont_idx] = (p, self.video_class(p, **self.video_class_kwargs))
            # c.create_underlying_cont(self.gpu_id)
        _, c = self.conts[cont_idx]
        assert c is not None
        idx = i - self.fs_cumsum[cont_idx]
        ret = c[idx]

        labels = None
        if self.labels_fn:
            labels = self.labels_fn(
                p,
                ret["frame_start_idx"],
                ret["frame_end_idx"],
            )
            ret.update(labels)

        if self.transform_fn is not None:
            ret = self.transform_fn(ret)
        return ret

    def __len__(self):
        return self.fs_cumsum[-1]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Information about contributing universities.
"""

UNIV_TO_BUCKET = {
    "bristol": "ego4d-bristol",
    "cmu": "ego4d-cmu",
    "frl_track_1_public": "ego4d-consortium-sharing",
    "georgiatech": "ego4d-georgiatech",
    "iiith": "ego4d-iiith",
    "indiana": "ego4d-indiana",
    "kaust": "ego4d-kaust",
    "minnesota": "ego4d-minnesota",
    "nus": "ego4d-speac",
    "sfu": "ego4d-sfu",
    "unc": "ego4d-unc",
    "unict": "ego4d-unict-milan",
    "utokyo": "ego4d-utokyo",
    "uniandes": "ego4d-university-sa",
    "cmu_africa": "ego4d-universityaf",
    "upenn": "ego4d-penn",
    "fair": "ego4d-fair",
}

BUCKET_TO_UNIV = {v: k for k, v in UNIV_TO_BUCKET.items()}



#!/usr/bin/env python3
"""
taken from
https://github.com/marcel-dancak/lz-string-python/blob/master/lzstring.py
"""
import math

KEYSTRURISAFE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$"
base_reverse_dict = {}


class Object:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def get_base_value(alphabet, character):
    if alphabet not in base_reverse_dict:
        base_reverse_dict[alphabet] = {}
    for i in range(len(alphabet)):
        base_reverse_dict[alphabet][alphabet[i]] = i
    return base_reverse_dict[alphabet][character]


def decompress(length, reset_value, get_next_value):
    dictionary = {}
    enlarge_in = 4
    dict_size = 4
    num_bits = 3
    entry = ""
    result = []

    data = Object(val=get_next_value(0), position=reset_value, index=1)

    for i in range(3):
        dictionary[i] = i

    bits = 0
    maxpower = math.pow(2, 2)
    power = 1

    while power != maxpower:
        resb = data.val & data.position
        data.position >>= 1
        if data.position == 0:
            data.position = reset_value
            data.val = get_next_value(data.index)
            data.index += 1

        bits |= power if resb > 0 else 0
        power <<= 1

    next = bits
    if next == 0:
        bits = 0
        maxpower = math.pow(2, 8)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = reset_value
                data.val = get_next_value(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1
        c = chr(bits)
    elif next == 1:
        bits = 0
        maxpower = math.pow(2, 16)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = reset_value
                data.val = get_next_value(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1
        c = chr(bits)
    elif next == 2:
        return ""

    dictionary[3] = c
    w = c
    result.append(c)
    counter = 0
    while True:
        counter += 1
        if data.index > length:
            return ""

        bits = 0
        maxpower = math.pow(2, num_bits)
        power = 1
        while power != maxpower:
            resb = data.val & data.position
            data.position >>= 1
            if data.position == 0:
                data.position = reset_value
                data.val = get_next_value(data.index)
                data.index += 1
            bits |= power if resb > 0 else 0
            power <<= 1

        c = bits
        if c == 0:
            bits = 0
            maxpower = math.pow(2, 8)
            power = 1
            while power != maxpower:
                resb = data.val & data.position
                data.position >>= 1
                if data.position == 0:
                    data.position = reset_value
                    data.val = get_next_value(data.index)
                    data.index += 1
                bits |= power if resb > 0 else 0
                power <<= 1

            dictionary[dict_size] = chr(bits)
            dict_size += 1
            c = dict_size - 1
            enlarge_in -= 1
        elif c == 1:
            bits = 0
            maxpower = math.pow(2, 16)
            power = 1
            while power != maxpower:
                resb = data.val & data.position
                data.position >>= 1
                if data.position == 0:
                    data.position = reset_value
                    data.val = get_next_value(data.index)
                    data.index += 1
                bits |= power if resb > 0 else 0
                power <<= 1
            dictionary[dict_size] = chr(bits)
            dict_size += 1
            c = dict_size - 1
            enlarge_in -= 1
        elif c == 2:
            return "".join(result)

        if enlarge_in == 0:
            enlarge_in = math.pow(2, num_bits)
            num_bits += 1

        if c in dictionary:
            entry = dictionary[c]
        else:
            if c == dict_size:
                entry = w + w[0]
            else:
                return None
        result.append(entry)

        # Add w+entry[0] to the dictionary.
        dictionary[dict_size] = w + entry[0]
        dict_size += 1
        enlarge_in -= 1

        w = entry
        if enlarge_in == 0:
            enlarge_in = math.pow(2, num_bits)
            num_bits += 1


def decompress_from_encoded_uri(compressed):
    if compressed is None:
        return ""
    if compressed == "":
        return None
    compressed = compressed.replace(" ", "+")
    decompressed = decompress(
        len(compressed),
        32,
        lambda index: get_base_value(KEYSTRURISAFE, compressed[index]),
    )

    return decompressed



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from setuptools import find_packages, setup

setup(
    name="ego4d",
    version="1.7.3",
    author="FAIR",
    author_email="info@ego4d-data.org",
    description="Ego4D Dataset CLI",
    url="https://github.com/facebookresearch/Ego4d/",
    install_requires=[
        "boto3",
        "tqdm",
        "regex",
        "dataclasses_json",
        "iopath",
    ],
    tests_require=[
        "pytest",
        "moto",
    ],
    packages=find_packages(exclude=("tests", "tests.*")),
    entry_points={
        "console_scripts": [
            "ego4d=ego4d.cli.cli:main",
            "egoexo=ego4d.egoexo.download.cli:main",
            "egoexo_internal=ego4d.internal.download.cli:internal_main",
            "ego4d_validation = ego4d.internal.validation.cli:main",
        ],
    },
)



import functools
import os
import random
import sys

import h5py
import pandas as pd
import torch
from ego4d.features.config import load_model
from ego4d.features.inference import _load_kinetics_class_names
from ego4d.research.clep.config import K400PreprocessConfig, TrainConfig
from ego4d.research.clep.preprocess.common import (
    get_language_model,
    run_feature_extraction,
)
from ego4d.research.common import batch_it, create_executor
from omegaconf import OmegaConf
from tqdm.auto import tqdm


def _preprocess_k400_data(video_path_label_pairs, feature_extract_config):
    model = load_model(feature_extract_config, patch_final_layer=True)

    ret = {}
    for path, label in tqdm(video_path_label_pairs):
        predictions = run_feature_extraction(path, model, feature_extract_config)
        if predictions is None:
            continue

        ret[path] = {
            "features": predictions.result[path],
            "label": label,
        }
    return ret


def preprocess_k400_data(config: TrainConfig, k_config: K400PreprocessConfig):
    """
    Assumptions:
        - input split given in a CSV file of the form:
            - <csv_dir>/<set_to_use>.csv
        - videos given in directory of the form:
            - "<youtube_id>_<start_time>_<end_time>.mp4
            - start_time / end_time are padded with up to 6 0's, e.g. --07WQ2iBlw_000001_000011.mp4
    """
    random.seed(1337)

    out_dir = os.path.join(config.pre_config.root_dir, k_config.root_dir)
    os.makedirs(out_dir, exist_ok=True)

    # TODO: configure
    val_set = pd.read_csv(os.path.join(k_config.csv_dir, f"{k_config.set_to_use}.csv"))

    def process_label(label):
        ret = label.replace('"', "")
        ret = ret.replace("'", "")
        ret = ret.replace("_", " ")
        return ret

    idx_to_label = list(_load_kinetics_class_names().items())
    idx_to_label.sort(key=lambda x: x[0])
    idx_to_label = dict(idx_to_label)
    assert sorted(idx_to_label.keys()) == list(idx_to_label.keys())

    label_names = list(idx_to_label.values())
    label_to_idx = {v: k for k, v in idx_to_label.items()}
    sentences = [
        f"The person in this video is doing {process_label(label)}"
        for _, label in idx_to_label.items()
    ]
    video_path_label_pairs = [
        (
            f"{k_config.dataset_dir}/{row.youtube_id}_{row.time_start:06d}_{row.time_end:06d}.mp4",
            row.label,
        )
        for row in val_set.itertuples()
    ]

    old_len = len(video_path_label_pairs)
    video_path_label_pairs = [
        val for val in video_path_label_pairs if os.path.exists(val[0])
    ]
    print(f"{old_len} -> {len(video_path_label_pairs)} examples", flush=True)

    feature_extract_config = OmegaConf.load(
        config.input_config.feature_extract_config_path
    )
    map_fn = functools.partial(
        _preprocess_k400_data,
        feature_extract_config=feature_extract_config,
    )
    batches = batch_it(
        video_path_label_pairs, batch_size=k_config.num_labels_per_machine
    )

    label_name_pairs = []
    if config.run_locally:
        raise AssertionError("not supported yet")
    else:
        slurm_config = config.pre_config.slurm_config
        print(
            f"To schedule {len(batches)} batches across {slurm_config.slurm_array_parallelism} machines"
        )
        cont = input("Continue? [y/N]: ")
        if cont != "y":
            print("Exiting...")
            sys.exit(0)
        executor = create_executor(slurm_config, len(batches))
        jobs = executor.map_array(map_fn, batches)

        # wait for the results
        out_path = os.path.join(out_dir, k_config.viz_feature_path)
        with h5py.File(out_path, "w") as out_f:
            for job in tqdm(jobs):
                vs = job.result()
                for k, v in vs.items():
                    label_name_pairs.append((k, v["label"]))
                    out_f.create_dataset(k, data=v["features"].numpy())

    # sentences
    print("Processing labels as sentences", flush=True)
    meta = {
        "labels": label_name_pairs,
        "label_text": [],
        "label_fv": [],
        "label_name_to_idx": label_to_idx,
        "idx_to_label_name": idx_to_label,
    }
    model = get_language_model(config)
    for idx, (sent, label_name) in tqdm(
        enumerate(zip(sentences, label_names)), total=len(sentences)
    ):
        assert label_to_idx[label_name] == idx
        fv = model.encode(sent, show_progress_bar=False)
        meta["label_text"].append(sent)
        meta["label_fv"].append(fv)

    out_meta_path = os.path.join(out_dir, k_config.metadata_out_path)
    torch.save(meta, out_meta_path)



"""
Performs downscaling of takes videos with the SLURM cluster.

Credits to Santhosh Kumar Ramakrishnan for providing the original snippet this
code originates from.
"""

import datetime
import json
import math
import os
import subprocess as sp
from concurrent.futures import ThreadPoolExecutor

import submitit
from ego4d.research.common import batch_it
from tqdm.auto import tqdm


ROOT_DIR = "/large_experiments/egoexo/v2/"
DS_TAKES_DIR = "/checkpoint/miguelmartin/egoexo/v2/downscaled_takes/takes_by_uni"


def call_ffmpeg(paths):
    src_path, tgt_path = paths
    assert os.path.exists(src_path)
    # https://docs.nvidia.com/video-technologies/video-codec-sdk/12.0/ffmpeg-with-nvidia-gpu/index.html
    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-loglevel",
        "error",
        # TODO: try with cuda encoding for faster processing
        # '-hwaccel', 'cuda',
        # '-hwaccel_output_format', 'cuda',
        "-i",
        src_path,
        # This sweet conditional is thanks to ChatGPT :)
        "-vf",
        "scale=w=if(lt(iw\,ih)\,448\,-2):h=if(lt(iw\,ih)\,-2\,448)",  # noqa
        # '-c:a', 'copy',
        # '-c:v', 'h264_nvenc',
        # '-b:v', '5M',
        tgt_path,
        "-y",
    ]
    print(" ".join(cmd))
    os.makedirs(os.path.dirname(tgt_path), exist_ok=True)
    sp.call(cmd)


def process_all(paths):
    map_fn = call_ffmpeg
    with ThreadPoolExecutor(5) as pool:
        for _ in tqdm(
            pool.map(map_fn, paths), total=len(paths), desc="Processing takes"
        ):
            continue


def main():
    num_machines: int = 50
    root_dir: str = ROOT_DIR
    ds_take_dir: str = DS_TAKES_DIR

    takes_to_process = json.load(open(os.path.join(root_dir, "takes.json")))

    map_values = []
    completed = 0
    num_vids = 0
    for take in takes_to_process:
        for _, streams in take["frame_aligned_videos"].items():
            for _, stream in streams.items():
                rel_path = stream["relative_path"]
                if rel_path is None:
                    continue
                if stream["is_collage"]:
                    continue
                src_path = os.path.join(root_dir, take["root_dir"], rel_path)
                dst_path = os.path.join(
                    ds_take_dir, take["university_id"], take["take_name"], rel_path
                )
                assert os.path.exists(src_path), src_path
                num_vids += 1
                if os.path.exists(dst_path):
                    completed += 1
                    continue
                map_values.append((src_path, dst_path))

    print(
        f"# videos to process: {len(map_values)} / {num_vids} [{completed} / {1 - (len(map_values) / num_vids):.2%} completed]"
    )
    job_inputs = batch_it(
        map_values, batch_size=math.ceil(len(map_values) / num_machines)
    )
    num_machines = min(num_machines, len(job_inputs))

    dt_now = datetime.datetime.now().strftime("%y%m%d_%H%M")
    log_dir = f"downscale/{dt_now}"
    print(f"Logging to: {log_dir}")
    os.makedirs(log_dir, exist_ok=True)
    executor = submitit.AutoExecutor(folder=log_dir)

    executor.update_parameters(
        timeout_min=3200,
        slurm_array_parallelism=num_machines,
        slurm_constraint="volta",
        slurm_partition="eht",
        gpus_per_node=1,
        cpus_per_task=10,
    )
    jobs = executor.map_array(process_all, job_inputs)

    print("Waiting...")
    results = []
    for job in tqdm(jobs):
        results.append(job.result())
    print("Done")


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass

import torch
from ego4d.features.config import BaseModelConfig, InferenceConfig
from maws.model_builder import build_model
from pytorchvideo.transforms import ApplyTransformToKey
from torch.nn import Module
from torchvision.transforms import CenterCrop, Compose, Lambda, Normalize, Resize


@dataclass
class ModelConfig(BaseModelConfig):
    model_name: str = "vit_2b14_xlmr_l"
    base_model: str = "maws_clip"
    input_type: str = "video"


class WrapModel(Module):
    def __init__(self, model: Module):
        super().__init__()
        self.model = model

    def forward(self, x) -> torch.Tensor:
        imgs = x["video"].half()
        imgs = imgs.view(-1, 3, imgs.shape[-2], imgs.shape[-1])
        return self.model.encode_images(imgs)


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    assert patch_final_layer, "maws does not provide a head"

    model = build_model(config.model_name, config.base_model)
    # model_name = f"{config.model_name}_{config.base_model}"
    # model = torch.hub.load("facebookresearch/maws", model=model_name)

    # Set to GPU or CPU
    model = WrapModel(model)
    model = model.to(inference_config.device)
    model = model.eval().half()
    return model


def norm_pixels(x):
    return x / 255.0


def video_to_image(x):
    x = x.permute(1, 0, 2, 3).squeeze(0)
    return x


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    assert inference_config.frame_window == 1
    transforms = [
        Lambda(norm_pixels),
        Lambda(video_to_image),
        Resize(size=224, interpolation=3),  # pyre-ignore
        CenterCrop(size=224),
        Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]

    return ApplyTransformToKey(
        key="video",
        transform=Compose(transforms),
    )



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import functools
import time
from typing import List, Optional, Tuple

import hydra
import torch
from ego4d.features.config import FeatureExtractConfig, get_videos, load_model, Video
from ego4d.features.extract_features import extract_features, num_fvs
from ego4d.features.slurm import create_executor


def profile_extraction(config: FeatureExtractConfig):
    videos, all_videos = get_videos(config)
    if videos:
        print(
            f"Video 0/{len(videos)} ({videos[0].frame_count} | {videos[0].path}): {videos[0]}"
        )
    # videos = [v for v in all_videos if v.frame_count <= 500]
    videos = [v for v in all_videos if v.frame_count > 2000 and v.frame_count <= 10000]
    # videos = [v for v in all_videos if "f1bdf9f3-4f65-4c70-b8ba-b3d4607c0cff" in v.uid]
    videos = videos[1:2]
    print(videos)

    assert len(videos) > 0, "No videos to process!"

    print(f"Got {len(videos)} videos")

    batch_sizes = [1]
    num_workers = [2]
    prefetch_factor = [2]
    model = load_model(config)

    num_examples = -1

    print(
        "prefetch_factor,batch_size,num_workers,total,mean,forward_pass,to_load,transfer_to_device"  # noqa
    )
    for batch_size, nw, pf in zip(batch_sizes, num_workers, prefetch_factor):
        config.inference_config.batch_size = batch_size
        config.inference_config.num_workers = nw
        config.inference_config.prefetch_factor = pf

        t1 = time.time()
        time_stats = extract_features(
            videos=videos,
            config=config,
            model=model,
            log_info=False,
            max_examples=num_examples,
            silent=False,
        ).time_stats
        t2 = time.time()

        total_time = t2 - t1

        forward_pass = torch.Tensor(time_stats.forward_pass)
        to_load = torch.Tensor(time_stats.to_load)
        transfer_time = torch.Tensor(time_stats.transfer_device)

        assert len(forward_pass.shape) == 1
        assert len(to_load.shape) == 1
        assert len(transfer_time.shape) == 1

        mean_sum = forward_pass.mean() + to_load.mean() + transfer_time.mean()
        mean_sum /= max(1, batch_size)

        if num_examples > 0:
            assert forward_pass.shape[0] * batch_size == num_examples

        if batch_size == 0:
            print(
                f"{pf},{batch_size},{nw},{total_time},{mean_sum},{forward_pass.mean()},{to_load.mean()},{transfer_time.mean()}"  # noqa
            )
        else:
            print(
                f"{pf},{batch_size},{nw},{total_time},{mean_sum},{forward_pass.mean()/batch_size},{to_load.mean()/batch_size},{transfer_time.mean()/batch_size}"  # noqa
            )


@hydra.main(config_path="configs", config_name=None)
def schedule_profile_extraction(config: FeatureExtractConfig):
    if config.schedule_config.run_locally:
        profile_extraction(config)
    else:
        executor = create_executor(config.schedule_config)
        job = executor.submit(functools.partial(profile_extraction, config=config))
        print(f"{job}")

        # wait for the job
        job.result()


if __name__ == "__main__":
    schedule_profile_extraction()  # pyre-ignore



import copy

import numpy as np

import torch.nn as nn
from ego4d.features.config import FeatureExtractConfig, Video
from ego4d.features.extract_features import extract_features
from ego4d.features.inference import _video_info as video_info
from ego4d.research.clep.config import TrainConfig
from sentence_transformers import SentenceTransformer


def run_feature_extraction(
    path: str, model: nn.Module, feature_extract_config: FeatureExtractConfig
):
    v_info = video_info(path)
    # pyre-ignore
    vid = Video(
        path,
        path,
        v_info["num_frames"],
        w=None,
        h=None,
        has_audio=False,
        is_stereo=False,
    )
    if vid.frame_count is None:
        return None

    feature_extract_config = copy.deepcopy(feature_extract_config)
    fps = v_info["fps"]
    assert fps is not None
    feature_extract_config.inference_config.fps = int(np.round(float(fps)))
    feature_extract_config.inference_config.stride = int(
        np.round(float(fps * (16 / 30)))
    )
    feature_extract_config.inference_config.frame_window = int(
        np.round(float(fps * (32 / 30)))
    )

    return extract_features(
        videos=[vid],
        config=feature_extract_config,
        model=model,
        log_info=False,
        silent=True,
        assert_feature_size=False,
    )


def get_language_model(config: TrainConfig) -> nn.Module:
    return SentenceTransformer(config.pre_config.ego4d_narr.st_model_name)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import importlib
import json
import os
import random
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import hydra
import pandas as pd
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf
from torchaudio.transforms import Resample
from torchvision.transforms import Compose, Lambda


@dataclass
class Video:
    """
    Description of a video
    """

    uid: str
    path: str
    frame_count: int
    w: int
    h: int
    has_audio: bool
    is_stereo: bool = False

    @property
    def dim(self) -> int:
        return (self.w * self.h) / (2 if self.is_stereo else 1)


@dataclass
class NormalizationConfig:
    normalize_audio: bool = False
    resample_audio_rate: int = 16000
    resampling_method: str = "sinc_interpolation"


@dataclass
class InputOutputConfig:
    """
    Assumptions:
    1. videos are in a single subdirectory with the pattern: <dir>/<uid>.mp4
    2. the manifest file exists in the same subdirectory

    If you need to adjust any of these assumptions, then please refer to:
    - `_path_for`, and
    - `_uid_to_frame_count`
    """

    # output
    out_path: str

    # input
    # TODO: add file list
    filter_completed: bool = True
    video_dir_path: str = "/datasets01/ego4d_track2/v1/full_scale/"
    ego4d_download_dir: str = "/checkpoint/miguelmartin/ego4d/"
    dataset_version: str = "ego4d"
    egoexo_data_dir: str = "/checkpoint/miguelmartin/egoexo_data/dev/"
    uid_list: Optional[List[str]] = None
    video_limit: int = -1
    debug_mode: bool = False
    debug_path: Optional[str] = None
    exclude_no_audio: bool = False
    eligible_cam_prefixes: Optional[List[str]] = None


@dataclass
class InferenceConfig:
    device: str = "cuda"

    video_reader_class: str = "PyAvReader"
    video_reader_kwargs_override: dict = field(default_factory=lambda _: {})

    # 0 == don't use dataloader
    # >0 use dataloader with bs=batch_size
    batch_size: int = 1

    # only used if batch_size != 0
    num_workers: int = 9

    prefetch_factor: int = 2

    fps: int = 30
    frame_window: int = 32
    stride: int = 16
    include_audio: bool = False
    include_video: bool = True
    norm_config: NormalizationConfig = field(default_factory=NormalizationConfig)


@dataclass
class ScheduleConfig:
    run_locally: bool = False

    log_folder: str = "slurm_log/%j"

    # Scheduler Configuration
    timeout_min: int = int(12 * 60)
    constraint: str = "volta"
    slurm_partition: str = "pixar"
    slurm_array_parallelism: int = 256
    gpus_per_node: int = 1
    cpus_per_task: int = 10

    # Batching Configuration
    overhead: float = 2  # off in the worst case -- estimate will be wrong
    time_per_forward_pass: float = 0.8

    schedule_time_per_node: float = 10


@dataclass
class BaseModelConfig:
    pass


@dataclass(order=True)
class FeatureExtractConfig:
    io: InputOutputConfig
    inference_config: InferenceConfig
    schedule_config: ScheduleConfig
    model_config: BaseModelConfig
    model_module_str: str = "ego4d.features.models.slowfast"
    force_yes: bool = False
    check_fv_count: bool = True


def get_model_module(config: FeatureExtractConfig):
    return importlib.import_module(config.model_module_str)


def _uids_for_dir(path: str) -> List[str]:
    if not os.path.exists(path):
        return []
    ret = [
        p
        for p in os.listdir(path)
        if Path(p).suffix not in [".json", ".csv", ".csv"]
        and not p.startswith(".")
        and not p.startswith("manifest")
    ]
    return [Path(p).stem for p in ret]


def _path_for(config: InputOutputConfig, uid: str) -> str:
    return f"{config.video_dir_path}/{uid}.mp4"


def _unfiltered_uids(config: InputOutputConfig) -> List[str]:
    uids = config.uid_list
    if uids is None:
        assert config.video_dir_path is not None, "Not given any uids"
        uids = _uids_for_dir(config.video_dir_path)
    return uids


def _uids(config: InputOutputConfig) -> List[str]:
    uids = _unfiltered_uids(config)

    if config.filter_completed:
        completed_uids = set(_uids_for_dir(config.out_path))
        uids = [uid for uid in uids if uid not in completed_uids]

    assert uids is not None, "`uids` is None"
    assert len(uids) >= 0, "`len(uids)` is 0"
    return uids


def _video_paths(config: InputOutputConfig, uids: List[str]) -> List[str]:
    return [_path_for(config, uid) for uid in uids]


def _uid_to_info(config: InputOutputConfig) -> Dict[str, int]:
    manifest_df = pd.read_csv(f"{config.video_dir_path}/manifest.csv")
    return {
        row.video_uid: {
            "num_frames": row.canonical_num_frames,
            "w": row.canonical_display_width,
            "h": row.canonical_display_height,
            "has_audio": not (
                pd.isnull(row.canonical_audio_start_sec)
                and pd.isnull(row.canonical_audio_duration_sec)
            ),
        }
        for row in manifest_df.itertuples()
    }


def _uid_to_is_stereo(config: InputOutputConfig) -> Dict[str, bool]:
    data_json = json.load(open(f"{config.ego4d_download_dir}/ego4d.json"))
    return {v["video_uid"]: v["is_stereo"] for v in data_json["videos"]}


def _videos(config: InputOutputConfig, unfiltered: bool = False) -> List[Video]:
    if config.dataset_version == "ego4d":
        uids = _uids(config) if not unfiltered else _unfiltered_uids(config)
        uid_to_info = _uid_to_info(config)
        uids_to_is_stereo = _uid_to_is_stereo(config)
        videos = [
            Video(
                uid=uid,
                path=_path_for(config, uid),
                frame_count=uid_to_info[uid]["num_frames"],
                w=uid_to_info[uid]["w"],
                h=uid_to_info[uid]["h"],
                has_audio=uid_to_info[uid]["has_audio"],
                is_stereo=uids_to_is_stereo[uid],
            )
            for uid in uids
            if uid in uid_to_info
        ]
        if config.exclude_no_audio:
            return [v for v in videos if v.has_audio]

        return videos
    else:
        takes = json.load(open(os.path.join(config.egoexo_data_dir, "takes.json")))
        all_uids = [t["take_uid"] for t in takes]
        uids = config.uid_list
        if uids is None:
            uids = all_uids
        if uids and takes:
            uid_takes = [t for t in takes if t["take_uid"] in uids]
            if len(uid_takes) < len(takes):
                print(f"Filtered {len(takes)} -> {len(uid_takes)} on uid config")
                takes = uid_takes
        completed_uids = set(_uids_for_dir(config.out_path))
        videos = []
        for take in takes:
            for cam_id, streams in take["frame_aligned_videos"].items():
                eligible_prefixes = config.eligible_cam_prefixes or [
                    "cam",
                    "aria",
                    "gp",
                ]
                if not any(x in cam_id.lower() for x in eligible_prefixes):
                    continue
                for stream_name, stream in streams.items():
                    # Config?
                    if "aria" in cam_id and stream_name != "rgb":
                        continue
                    is_aria = "aria" in stream["cam_id"]
                    # NOTE: known constants for not downsampled videos
                    w = 1408 if is_aria else 3840
                    h = 1408 if is_aria else 2160
                    uid = f"{take['take_uid']}_{cam_id}_{stream_name}"
                    if (
                        not unfiltered
                        and config.filter_completed
                        and uid in completed_uids
                    ):
                        continue

                    path = os.path.join(
                        config.egoexo_data_dir,
                        take["root_dir"],
                        stream["relative_path"],
                    )

                    if not os.path.exists(path):
                        continue

                    videos.append(
                        Video(
                            uid=uid,
                            path=path,
                            # NOTE: w/h used to estimate time to complete
                            w=w,
                            h=h,
                            frame_count=take["timesync_end_idx"]
                            - take["timesync_start_idx"]
                            - 1,
                            has_audio=False,
                            is_stereo=False,
                        )
                    )
        return videos


def get_videos(config: FeatureExtractConfig) -> Tuple[List[Video], List[Video]]:
    """
    Return (videos_to_process, all_videos)
    """
    possibly_filtered_videos = _videos(config.io, unfiltered=False)
    all_videos = _videos(config.io, unfiltered=True)
    if config.io.video_limit > 0:
        random.shuffle(possibly_filtered_videos)
        return possibly_filtered_videos[0 : config.io.video_limit], all_videos
    return possibly_filtered_videos, all_videos


def get_transform(config: FeatureExtractConfig) -> Any:
    ic = config.inference_config
    nc = ic.norm_config
    model_transform = get_model_module(config).get_transform(ic, config.model_config)
    transforms = []
    if hasattr(config, "norm_config") and config.norm_config.normalize_audio:
        print(f"Normalizing with: {config.norm_config}")

        def resample_audio(x):
            return Resample(
                orig_freq=x["audio_sample_rate"],
                new_freq=nc.resample_audio_rate,
                resampling_method=nc.resampling_method,
            )

        transforms += [Lambda(resample_audio)]

    transforms += [model_transform]
    return Compose(transforms)


def load_model(config: FeatureExtractConfig, patch_final_layer: bool = True) -> Any:
    module = get_model_module(config)
    return module.load_model(
        config.inference_config,
        config.model_config,
        patch_final_layer=patch_final_layer,
    )


@hydra.main(config_path="configs", config_name=None)
def test_load_config(config: FeatureExtractConfig):
    print(
        f"""
    Config:

{OmegaConf.to_yaml(config)}
    """
    )


if __name__ == "__main__":
    from ego4d.features.models.slowfast import ModelConfig

    cs = ConfigStore.instance()
    cs.store(
        name="default",
        node=FeatureExtractConfig(
            io=InputOutputConfig(),
            inference_config=InferenceConfig(),
            schedule_config=ScheduleConfig(),
            model_config=ModelConfig(),
        ),
    )

    test_load_config()  # pyre-ignore



# pyre-strict
import ast
import csv
import json
import os
from collections import defaultdict
from dataclasses import dataclass, fields
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from ego4d.internal.s3 import ls_relative

from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler

pathmgr = PathManager()
pathmgr.register_handler(S3PathHandler(profile="default"))


DEFAULT_DATE_FORMAT_STR = "%Y-%m-%d %H:%M:%S"
VALID_DATE_FORMAT_STRS = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d"]


class ErrorLevel(Enum):
    WARN = 0
    ERROR = 1


@dataclass
class Error:
    level: ErrorLevel
    uid: str
    type: str
    description: str


@dataclass
class Device:
    device_id: int
    name: str


@dataclass
class ComponentType:
    component_type_id: int
    name: str


@dataclass
class Scenario:
    scenario_id: int
    name: str
    included_in_release_1: bool
    is_ad_hoc: bool


@dataclass
class VideoMetadata:
    university_video_id: str
    university_video_folder_path: str
    number_video_components: int
    start_date_recorded_utc: datetime
    recording_participant_id: str
    device_id: int
    video_device_settings: dict
    physical_setting_id: str
    video_scenario_ids: list


@dataclass
class VideoComponentFile:
    university_video_id: str
    video_component_relative_path: str
    component_index: int
    is_redacted: bool
    start_date_recorded_utc: datetime
    compression_settings: dict
    includes_audio: bool
    component_metadata: dict
    deidentification_metadata: dict


@dataclass
class AuxiliaryVideoComponentDataFile:
    university_video_id: str
    component_index: int
    component_type_id: int
    video_component_relative_path: str


@dataclass
class Particpant:
    participant_id: str
    participant_metadata: dict


@dataclass
class SynchronizedVideos:
    video_grouping_id: str
    synchronization_metadata: dict
    associated_videos: dict


@dataclass
class PhysicalSetting:
    setting_id: str
    name: str
    associated_matterport_scan_path: str


@dataclass
class Annotations:
    university_video_id: str
    start_seconds: float
    end_seconds: float
    annotation_data: dict


@dataclass
class StandardMetadata:
    devices: Dict[str, Device]
    component_types: Dict[str, ComponentType]
    scenarios: Dict[str, Scenario]


@dataclass
class Manifest:
    videos: Dict[str, VideoMetadata]
    video_components: Dict[str, List[VideoComponentFile]]
    aux_components: Dict[str, List[AuxiliaryVideoComponentDataFile]]
    participants: Dict[str, Particpant]
    sync_videos: Dict[str, SynchronizedVideos]
    physical_setting: Dict[str, PhysicalSetting]
    annotations: Dict[str, List[Annotations]]


# new EgoExo types


@dataclass
class DeviceEgoExo:
    device_id: str
    name: str


@dataclass
class CaptureMetadataEgoExo:
    university_capture_id: str
    university_video_folder_path: str
    number_videos: int
    number_takes: int
    post_surveys_relative_path: str
    physical_setting_id: str
    start_date_recorded_utc: datetime
    additional_metadata: dict


@dataclass
class TakeMetadataEgoExo:
    university_capture_id: str
    take_id: str
    scenario_id: int
    is_narrated: bool
    is_dropped: bool
    take_start_seconds_aria: float
    object_ids: list
    recording_participant_id: str
    additional_metadata: dict


@dataclass
class VideoMetadataEgoExo:
    university_capture_id: str
    university_video_id: str
    number_video_components: int
    is_ego: str
    has_walkaround: bool
    includes_audio: bool
    device_type: str
    device_id: str
    video_device_settings: dict
    additional_metadata: dict
    is_redacted: bool


@dataclass
class VideoComponentFileEgoExo:
    university_capture_id: str
    university_video_id: str
    video_component_relative_path: str
    component_index: int
    is_redacted: bool


@dataclass
class ColmapMetadataEgoExo:
    university_capture_id: str
    colmap_configuration_id: str
    config_relative_path: str
    colmap_ran: bool
    was_inspected: bool
    is_final_configuration: bool
    version: str
    notes: str


@dataclass
class ObjectMetadataEgoExo:
    university_object_id: str
    object_name: str
    object_relative_path: str
    physical_setting_id: str
    additional_metadata: dict


@dataclass
class ParticipantMetadataEgoExo:
    participant_id: str
    scenario_id: int
    collection_date: datetime
    pre_survey_data: dict
    participant_metadata: dict


@dataclass
class PhysicalSettingEgoExo:
    setting_id: str
    name: str


@dataclass
class ExtraDataEgoExo:
    university_capture_id: str
    take_id: str
    annotation_data: dict


@dataclass
class ScenarioEgoExo:
    scenario_id: int
    name: str


@dataclass
class StandardMetadataEgoExo:
    devices: Dict[str, Device]
    scenarios: Dict[str, ScenarioEgoExo]


@dataclass
class ManifestEgoExo:
    captures: Dict[str, CaptureMetadataEgoExo]  # by capture id
    takes: Dict[str, List[TakeMetadataEgoExo]]  # by capture id
    videos: Dict[str, List[VideoMetadataEgoExo]]  # by capture id
    video_components: Dict[
        Tuple[str, str], List[VideoComponentFileEgoExo]
    ]  # by (capture_id, video_id)
    colmap: Optional[Dict[str, List[ColmapMetadataEgoExo]]]  # by capture id
    physical_setting: Dict[str, PhysicalSettingEgoExo]  # by setting id
    objects: Optional[Dict[str, ObjectMetadataEgoExo]]  # by object id
    participants: Optional[Dict[str, ParticipantMetadataEgoExo]]  # by participant id
    extra_data: Optional[List[ExtraDataEgoExo]]


def default_decode(value: str, datatype: type, name: str) -> Any:
    if datatype in (dict, defaultdict):
        if len(value) == 0:
            return {}
        return json.loads(value)
    elif datatype == list:
        if len(value) == 0:
            return []
        return json.loads(value)
    elif datatype == datetime:
        for date_format_str in VALID_DATE_FORMAT_STRS:
            if len(value) == 0:
                return None
            try:
                return datetime.strptime(value.split(".")[0], date_format_str)
            except ValueError:
                pass
    elif datatype == int:
        if len(value) == 0:
            return None
        if "." in value:
            x = value.split(".")
            assert len(x) == 2, f"invalid number: {x}"
            assert int(x[1]) == 0, f"fractional part must be 0 for integer type: {x}"
            return int(x[0])
        else:
            return int(value)
    elif datatype == bool:
        temp = value.strip().lower()
        if len(temp) == 0:
            return False

        if temp[0] == "f":
            temp = "F" + temp[1:]
        elif temp[0] == "t":
            temp = "T" + temp[1:]

        return ast.literal_eval(temp)
    elif datatype == str:
        if len(value) == 0:
            return None
        return value
    elif datatype in (int, float):
        if len(value) == 0:
            return None
        return ast.literal_eval(value)


def load_dataclass_dict_from_csv(
    input_csv_file_path: str,
    dataclass_class: type,
    dict_key_field: Union[str, List[str]],
    unique_per_key: bool,
    default_decode_fn: Callable[[str, type], Any] = default_decode,
) -> Dict[Any, List[Any]]:
    """
    Args:
        input_csv_file_path (str): File path of the csv to read from
        dataclass_class (type): The dataclass to read each row into.
        dict_key_field (str): The field of 'dataclass_class' to use as
            the dictionary key.
        list_per_key (bool) = False: If the output data structure
        contains a list of dataclass objects per key, rather than a
        single unique dataclass object.

    Returns:
        Dict[Any, Union[Any, List[Any]] mapping from the dataclass
        value at attr = dict_key_field to either:

        if 'list_per_key', a list of all dataclass objects that
        have equal values at attr = dict_key_field, equal to the key

        if not 'list_per_key', the unique dataclass object
        for which the value at attr = dict_key_field is equal to the key

    Raises:
        AssertionError: if not 'list_per_key' and there are
        dataclass objects with equal values at attr = dict_key_field
    """

    output = defaultdict(list)
    with pathmgr.open(input_csv_file_path) as csvfile:
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')  # pyre-ignore
        try:
            column_index = {header: i for i, header in enumerate(next(reader))}
        except StopIteration:
            # there's no column information
            return output

        if len(column_index) == 0:
            # there's no column information
            return output

        field_name_to_metadata = {
            f.name: {
                "decode_fn": (
                    f.metadata["csv_decode_fn"]
                    if "csv_decode_fn" in f.metadata
                    else default_decode_fn
                ),
                "column_index": column_index.get(f.name, None),
                "type": f.type,
            }
            for f in fields(dataclass_class)
        }
        csv_names = set(column_index.keys())
        dn_names = set(field_name_to_metadata.keys())
        if csv_names != dn_names:
            missing_str = "\n".join([f"- {x}" for x in (dn_names - csv_names)])
            additional_str = "\n".join([f"- {x}" for x in (csv_names - dn_names)])
            raise AssertionError(
                f"""
Missing fields from CSV {input_csv_file_path}:
{missing_str}
Additional fields in CSV:
{additional_str}
"""
            )

        duplicate_keys = []
        lineno = 2
        for line in reader:
            constructor_params = {}
            for name, meta in field_name_to_metadata.items():
                try:
                    constructor_params[name] = meta["decode_fn"](
                        line[meta["column_index"]],
                        meta["type"],
                        name,
                    )

                except Exception as e:
                    print(
                        f"Could not decode column: '{name}' for input file: {input_csv_file_path} Will not decode. Line: {lineno}"
                    )
                    raise e

            obj = dataclass_class(**constructor_params)

            dict_key = (
                getattr(obj, dict_key_field)
                if isinstance(dict_key_field, str)
                else tuple(getattr(obj, k) for k in dict_key_field)
            )
            output[dict_key].append(obj)
            if unique_per_key:
                if len(output[dict_key]) != 1:
                    duplicate_keys.append(dict_key)
            lineno += 1

        if len(duplicate_keys) > 0:
            raise AssertionError(
                f"Multiple Primary Keys/IDs found for file {input_csv_file_path}:\n{duplicate_keys}"
            )
    return output


def load_standard_metadata_files(
    standard_metadata_folder: str,
) -> StandardMetadata:
    if not pathmgr.exists(standard_metadata_folder):
        raise AssertionError(f"{standard_metadata_folder} does not exist")
    available_files = ls_relative(standard_metadata_folder, pathmgr)

    file_name = "device.csv"
    if file_name not in available_files:
        raise AssertionError(
            f"required file {file_name} not found in " f"{standard_metadata_folder}"
        )

    file_path = os.path.join(standard_metadata_folder, file_name)
    devices = load_dataclass_dict_from_csv(
        file_path,
        Device,
        "device_id",
        unique_per_key=True,
    )

    file_name = "component_type.csv"
    if file_name not in available_files:
        raise AssertionError(
            f"required file {file_name} not found in " f"{standard_metadata_folder}"
        )

    file_path = os.path.join(standard_metadata_folder, file_name)
    component_types = load_dataclass_dict_from_csv(
        file_path,
        ComponentType,
        "component_type_id",
        unique_per_key=True,
    )

    file_name = "scenario.csv"
    if file_name not in available_files:
        raise AssertionError(
            f"required file {file_name} not found in " f"{standard_metadata_folder}"
        )
    file_path = os.path.join(standard_metadata_folder, file_name)
    scenarios = load_dataclass_dict_from_csv(
        file_path,
        Scenario,
        "scenario_id",
        unique_per_key=True,
    )

    # TODO: fixme to updated stats
    assert len(scenarios) > 0
    assert len(devices) > 0
    assert len(component_types) > 0

    return StandardMetadata(
        devices={k: vs[0] for k, vs in devices.items()},
        component_types={k: vs[0] for k, vs in component_types.items()},
        scenarios={k: vs[0] for k, vs in scenarios.items()},
    )


def load_standard_metadata_files_egoexo(
    standard_metadata_folder: str,
) -> StandardMetadataEgoExo:
    if not pathmgr.exists(standard_metadata_folder):
        raise AssertionError(f"{standard_metadata_folder} does not exist")
    available_files = ls_relative(standard_metadata_folder, pathmgr)

    file_name = "device.csv"
    if file_name not in available_files:
        raise AssertionError(
            f"required file {file_name} not found in {standard_metadata_folder}"
        )

    file_path = os.path.join(standard_metadata_folder, file_name)
    devices = load_dataclass_dict_from_csv(
        file_path,
        DeviceEgoExo,
        "device_id",
        unique_per_key=True,
    )

    file_name = "scenario.csv"
    if file_name not in available_files:
        raise AssertionError(
            f"required file {file_name} not found in {standard_metadata_folder}"
        )
    file_path = os.path.join(standard_metadata_folder, file_name)
    scenarios = load_dataclass_dict_from_csv(
        file_path,
        ScenarioEgoExo,
        "scenario_id",
        unique_per_key=True,
    )

    # TODO: fixme to updated stats
    assert len(scenarios) > 0
    assert len(devices) > 0

    return StandardMetadataEgoExo(
        devices={k: vs[0] for k, vs in devices.items()},
        scenarios={k: vs[0] for k, vs in scenarios.items()},
    )


def load_egoexo_manifest(manifest_dir: str) -> ManifestEgoExo:
    available_files = ls_relative(manifest_dir, pathmgr)

    def _check_file_exists():
        if file_name not in available_files:
            raise AssertionError(
                f"required file {file_name} not found in {manifest_dir}"
            )

    file_name = "capture_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    capture_metadata = load_dataclass_dict_from_csv(
        file_path,
        CaptureMetadataEgoExo,
        "university_capture_id",
        unique_per_key=True,
    )

    file_name = "take_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    take_metadata = load_dataclass_dict_from_csv(
        file_path,
        TakeMetadataEgoExo,
        "university_capture_id",
        unique_per_key=False,
    )

    file_name = "video_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    video_metadata = load_dataclass_dict_from_csv(
        file_path,
        VideoMetadataEgoExo,
        "university_capture_id",
        unique_per_key=False,
    )

    file_name = "video_component_file.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    video_component_metadata = load_dataclass_dict_from_csv(
        file_path,
        VideoComponentFileEgoExo,
        ["university_capture_id", "university_video_id"],
        unique_per_key=False,
    )

    file_name = "colmap_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    colmap_metadata = None
    if file_name in available_files:
        colmap_metadata = load_dataclass_dict_from_csv(
            file_path,
            ColmapMetadataEgoExo,
            "university_capture_id",
            unique_per_key=False,
        )

    file_name = "object_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    object_metadata = None
    if file_name in available_files:
        object_metadata = load_dataclass_dict_from_csv(
            file_path,
            ObjectMetadataEgoExo,
            "university_object_id",
            unique_per_key=True,
        )

    file_name = "physical_setting.csv"
    file_path = os.path.join(manifest_dir, file_name)
    physical_setting = None
    if file_name in available_files:
        physical_setting = load_dataclass_dict_from_csv(
            file_path,
            PhysicalSettingEgoExo,
            "setting_id",
            unique_per_key=True,
        )

    file_name = "participant_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    participant_metadata = None
    if file_name in available_files:
        participant_metadata = load_dataclass_dict_from_csv(
            file_path,
            ParticipantMetadataEgoExo,
            ["participant_id", "scenario_id", "collection_date"],
            unique_per_key=True,
        )

    file_name = "extra_data.csv"
    file_path = os.path.join(manifest_dir, file_name)
    extra_data = None
    if file_name in available_files:
        extra_data = load_dataclass_dict_from_csv(
            file_path,
            ExtraDataEgoExo,
            "university_capture_id",
            unique_per_key=False,
        )

    return ManifestEgoExo(
        captures={k: v[0] for k, v in capture_metadata.items()},
        takes=take_metadata,
        videos=video_metadata,
        video_components=video_component_metadata,
        colmap=colmap_metadata,
        physical_setting=(
            {k: v[0] for k, v in physical_setting.items()}
            if physical_setting is not None
            else None
        ),
        objects=(
            {k: v[0] for k, v in object_metadata.items()}
            if object_metadata is not None
            else None
        ),
        participants=(
            {k: v[0] for k, v in participant_metadata.items()}
            if participant_metadata is not None
            else None
        ),
        extra_data=extra_data,  # pyre-ignore
    )


def load_manifest(manifest_dir: str) -> Manifest:
    available_files = ls_relative(manifest_dir, pathmgr)

    def _check_file_exists():
        if file_name not in available_files:
            raise AssertionError(
                f"required file {file_name} not found in {manifest_dir}"
            )

    file_name = "video_metadata.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    video_metadata_dict = load_dataclass_dict_from_csv(
        file_path,
        VideoMetadata,
        "university_video_id",
        unique_per_key=True,
    )

    # Load video_component_file.csv
    file_name = "video_component_file.csv"
    file_path = os.path.join(manifest_dir, file_name)
    _check_file_exists()
    video_components_dict = load_dataclass_dict_from_csv(
        file_path,
        VideoComponentFile,
        "university_video_id",
        unique_per_key=False,
    )

    # Load optional files
    # Load auxiliary_video_component_data_file.csv, if available
    file_name = "auxiliary_video_component_data_file.csv"
    auxiliary_video_component_dict = {}
    if file_name in available_files:
        file_path = os.path.join(manifest_dir, file_name)
        auxiliary_video_component_dict = load_dataclass_dict_from_csv(
            file_path,
            AuxiliaryVideoComponentDataFile,
            "university_video_id",
            unique_per_key=False,
        )

    # Load participant.csv, if available
    participant_dict = {}
    for file_name in [
        "participant_data.csv",
        "participant.csv",
        "participant_metadata.csv",
    ]:
        if file_name in available_files:
            file_path = os.path.join(manifest_dir, file_name)
            participant_dict = load_dataclass_dict_from_csv(
                file_path,
                Particpant,
                "participant_id",
                unique_per_key=True,
            )
            break

    # Load synchronized_videos.csv, if available
    file_name = "synchronized_videos.csv"
    synchronized_video_dict = {}
    if file_name in available_files:
        file_path = os.path.join(manifest_dir, file_name)
        synchronized_video_dict = load_dataclass_dict_from_csv(
            file_path,
            SynchronizedVideos,
            "video_grouping_id",
            unique_per_key=True,
        )

    # Load physical_setting.csv, if available
    file_name = "physical_setting.csv"
    physical_setting_dict = {}
    if file_name in available_files:
        file_path = os.path.join(manifest_dir, file_name)
        physical_setting_dict = load_dataclass_dict_from_csv(
            file_path,
            PhysicalSetting,
            "setting_id",
            unique_per_key=True,
        )

    # Load annotations.csv, if available
    file_name = "annotations.csv"
    annotations_dict = {}
    if file_name in available_files:
        file_path = os.path.join(manifest_dir, file_name)
        annotations_dict = load_dataclass_dict_from_csv(
            file_path,
            Annotations,
            "university_video_id",
            unique_per_key=False,
        )

    return Manifest(
        # unique keys
        videos={k: vs[0] for k, vs in video_metadata_dict.items()},
        participants={k: vs[0] for k, vs in participant_dict.items()},
        sync_videos={k: vs[0] for k, vs in synchronized_video_dict.items()},
        physical_setting={k: vs[0] for k, vs in physical_setting_dict.items()},
        # non-unique keys
        video_components=video_components_dict,
        aux_components=auxiliary_video_component_dict,
        annotations=annotations_dict,
    )


def load_released_video_files(
    released_video_path: Optional[str],
) -> Optional[Dict[str, List[str]]]:
    # TODO: changeme
    if released_video_path is None:
        return None

    with pathmgr.open(released_video_path) as csvfile:
        released_videos = defaultdict(list)
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        _ = next(reader)  # skip header
        for line in reader:
            released_videos[line[2]].append(line[0])
    return released_videos



import torch
from torch.nn import Module


class FeedVideoInput(Module):
    def __init__(self, model: Module):
        super().__init__()
        self.model = model

    def forward(self, x) -> torch.Tensor:
        return self.model(x["video"])
